#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>          
#include <sys/stat.h>      
#include <semaphore.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#define MAXITEMS 3
#define PIPENAME "roura"

typedef struct {
    int is_null;
    char value;
} itemtype;

sem_t *empty;
sem_t *full;
sem_t *mutex;


void produce(itemtype *item) {
    char *text ="Ahoooj!";
    static int i = 0;
    if(i != strlen(text)) {
        item->is_null = 0;
        item->value = text[i];
    }
    else {
        item->is_null = 1;
        item->value = '\0';
    }
    i++;
}

void enter_item(itemtype item,int pipe)
{
    write(pipe,&item, sizeof(itemtype));
    printf("Item Entered: { %i, %c }\n",item.is_null,item.value);
}

void remove_item(itemtype *item,int pipe)
{
    read(pipe,item, sizeof(itemtype));
}

int consume_item(itemtype item)
{
    printf("Item Consumed: { %i, %c }\n",item.is_null,item.value);
    sleep(2);
    if(item.is_null)
        return 1;
    return 0;
}

void producer(void)
{
    itemtype item;
    int pipe;
    if((pipe = open(PIPENAME, O_WRONLY)) == -1)
    {
        perror("Open write pipe:");
        exit(1);
    }
    while(1) {
        produce(&item);
        sem_wait(empty);
     //   sem_wait(mutex); 
        enter_item(item,pipe);
     //   sem_post(mutex);
        sem_post(full);
        if(item.is_null)
           break;
    }
    close(pipe);
}

void consumer(void) 
{
    itemtype item;
    int pipe;
    if((pipe = open(PIPENAME, O_RDONLY)) == -1)
    {
        perror("Open read pipe:");
        exit(1);
    }
    while (1) {
        sem_wait(full);
     //   sem_wait(mutex);
        remove_item(&item,pipe);
     //   sem_post(mutex);
        sem_post(empty);
        if(consume_item(item))
            break;
    }
    close(pipe);
}

int main(void)
{
    pid_t pid;
    int status;
    empty = sem_open("empty", O_CREAT, 0, MAXITEMS);
    if (empty == SEM_FAILED) {
        perror("Error creating semaphore empty:");
        exit(1);
    }
    full = sem_open("full",O_CREAT,0,0);
    if (full == SEM_FAILED) {
        perror("Error creating semaphore full:");
        exit(1);
    }
    mutex = sem_open("mutex",O_CREAT,0,1);
    if (mutex == SEM_FAILED) {
        perror("Error creating semaphore mutex:");
        exit(1);
    }
    mkfifo(PIPENAME, S_IWUSR | S_IRUSR | S_IRGRP);

    if(!(pid = fork())) {
        producer();
        printf("producer end\n");
    } else {
        printf("pid = %i\n",pid);
        consumer();
        printf("consumer end\n");
        waitpid(pid,&status,0);
        sem_close(empty);
        sem_close(full);
        sem_close(mutex);
    }
    return 0;
}
